#!/usr/bin/env python3
# *-* coding: utf-8 *-*
"""
Usage: unpack.py [options] -c CRX_FILE [EXT_DIR] [PASSWD]
       unpack.py [options] xo ZIP_PATH EXT_DIR [PASSWD]
       unpack.py [options] convert BASE_DIR [-s SKIP_FORMAT ...]

Options:
 -c CRX_FILE  The extension file to extract.
 -f           Force extraction. Deletes the destination folder if it already
              exists.
 -t           Turns off testing the zip file portion before attempting
              extraction.
 -g           Don't skip converting GIFs when converting images to PNG format.
              By default, GIFs will be skipped because they cannot always be
              encoded to PNG format.
 -s SKIP_FORMAT ...
              Do not attempt to convert the given list of image formats. If
              multiple formats are given, the -s flag must precede each,
              e.g. -s ICO -s WEBP. [Default: ICO WEBP PNG]

 --log-file=LOG_FILE
 --log-level=LOG_LEVEL
 --log-fmt=LOG_FMT
              Set the parameters for logging. Only has an effect when used
              with the `convert` option.

Unpacks the CRX_FILE by taking off the headers and extracting the zip file in
the same directory as the CRX_FILE. Optionally, unpacks to different directory
if EXT_DIR is specified.

With the `xo` command, only the extraction operation will take place. This
means the CRX header won't be parsed or tested, images won't be converted, and
the files' modes won't be changed.

With the `convert` command, all images in the BASE_DIR will be converted to
PNG format, except those specified with `-s`.
"""

import logging

from docopt import docopt
from PIL import Image

import crx_unpack
from crx_unpack.clr import add_color_log_levels


def main():
    args = docopt(__doc__)

    # Make sure all of the specified formats is an accepted format
    Image.init()  # Make sure the list of supported types has been initialized
    _fs = []
    for _f in args['-s']:
        _f = _f.upper()
        if _f in Image.OPEN:
            _fs.append(_f)
    args['s'] = _fs.copy()
    # Always skip PNG files
    if 'PNG' not in args['-s']:
        args['-s'].append('PNG')

    if args['xo']:
        # Only do the extraction
        crx_unpack.extract_zip(args['ZIP_PATH'], args['EXT_DIR'], pwd=args['PASSWD'], test_contents=(not args['-t']),
                               reraise_errors=not args['xo'])

    elif args['convert']:
        # Only do the image conversion for a specific directory

        # Set up logging
        kw = {}
        if args['--log-file'] is not None:
            kw['filename'] = args['--log-file']
        if args['--log-level'] is not None:
            kw['level'] = int(args['--log-level'])
        if args['--log-fmt'] is not None:
            kw['format'] = args['--log-fmt']
        logging.basicConfig(**kw)
        add_color_log_levels(center=True)

        # Start the conversion
        crx_unpack.convert_imgs(args['BASE_DIR'], skip_gifs=(not args['-g']), skip_other=args['-s'])

    else:
        _ext_dir = crx_unpack.unpack(args['-c'], args['EXT_DIR'], args['-f'], test_contents=(not args['-t']),
                                     passwd=args['PASSWD'], skip_img_formats=args['-s'])
        print("CRX extracted to: %s" % _ext_dir)


if __name__ == '__main__':
    main()



"""
Color text. Typical usage:

    red('red text', False)

Returns the string "red text" where the text will be red and the
background will be the default.

    red('red background')

Returns the string "red background" where the text will be the default
color and the background will be red.
"""

import logging
from colorama import init, Back, Fore  # BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET
init(autoreset=True)

__all__ = ['black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white']


def _color_it(text, color, bg):
    return getattr(bool(bg) and Back or Fore, color) + str(text) + getattr(bool(bg) and Back or Fore, 'RESET')


def black(text, background=True):
    return _color_it(text, 'BLACK', background)


def red(text, background=True):
    return _color_it(text, 'RED', background)


def green(text, background=True):
    return _color_it(text, 'GREEN', background)


def yellow(text, background=True):
    return _color_it(text, 'YELLOW', background)


def blue(text, background=True):
    return _color_it(text, 'BLUE', background)


def magenta(text, background=True):
    return _color_it(text, 'MAGENTA', background)


def cyan(text, background=True):
    return _color_it(text, 'CYAN', background)


def white(text, background=True):
    return _color_it(text, 'WHITE', background)


def add_color_log_levels(center=False):
    if center:
        c = 'CRITICAL'.center(8)
        e = 'ERROR'.center(8)
        w = 'WARNING'.center(8)
        i = 'INFO'.center(8)
        d = 'DEBUG'.center(8)
        n = 'NOTSET'.center(8)
    else:
        c = 'CRITICAL'
        e = 'ERROR'
        w = 'WARNING'
        i = 'INFO'
        d = 'DEBUG'
        n = 'NOTSET'
    logging.addLevelName(50, black(red(c, True)))
    logging.addLevelName(40, black(magenta(e, True)))
    logging.addLevelName(30, black(yellow(w, True)))
    logging.addLevelName(20, black(blue(i, True)))
    logging.addLevelName(10, black(green(d, True)))
    logging.addLevelName(0, black(white(n, True)))



# *-* coding: utf-8 *-*
"""
The purpose of this module is to mimic how Google Chrome unpacks CRX files
as closely as possible. Involved in this is the need to remove the CRX headers
(see the structure details of CRXs on the :doc:`Home <index>` page), separate
the underlying ZIP file, extract the contents of the ZIP file, among other
things.

For end users, the only function you should need to call is :py:func:`unpack`,
which will handle each of the steps mentioned above.
"""

import codecs
import logging
import os
import re
from os import path
from shutil import rmtree
from struct import Struct
from subprocess import check_call, CalledProcessError
from tempfile import NamedTemporaryFile
from zipfile import ZipFile, BadZipFile

try:
    from PIL import Image
except ImportError:
    FORBID_IMG_CONVERT = True

    # Just in case PIL isn't available *and* someone calls convert_imgs() directly...
    class Image:
        @staticmethod
        def open():
            raise OSError
else:
    FORBID_IMG_CONVERT = False

__all__ = ['unpack', 'BadCrxHeader', 'BadZipFile']

with open(path.abspath(path.join(path.dirname(__file__), 'VERSION'))) as _v:
    __version__ = _v.read().strip()

HEADER_FMT = Struct('<4s3I')
CONVERT_IMAGES = False
ERR_TYPE = (None,
            BadZipFile,
            MemoryError,
            IndexError,
            IsADirectoryError,
            NotADirectoryError)
DIR_MODE = 700
FILE_MODE = 644
JUST_EXIT = False


class BadCrxHeader(Exception):
    """Raised when a CRX's header length or values aren't valid."""


def unpack(crx_file, ext_dir=None, overwrite_if_exists=False, img_tallies=None, test_contents=True, passwd=None,
           skip_img_formats=None, unpack_in_subprocess=False, convert_in_subprocess=True, do_convert=CONVERT_IMAGES):
    """Unpack the CRX and extract it in the directory at ext_dir.

    Return the absolute, normalized path to the extraction directory (useful
    if it wasn't given as a parameter).

    :param str crx_file: Path to the CRX file.
    :param str ext_dir: Directory where to extract the contents.
    :param bool overwrite_if_exists: When extracting to a directory that already
        exists, unpack will normally fail. Setting this to True will delete
        the contents of the destination directory before unzipping.
    :param dict img_tallies: A dictionary for storing the number of each type of
        image file converted during the unpacking process.
    :param bool test_contents: When unpacking the CRX, use the zipfile module's
        test feature to test the validity of the embedded zip file before
        extraction.
    :param str passwd: Optional password to use when extracting the CRX. If the
        CRX was obtained from Google's `Chrome Web Store
        <https://chrome.google.com/webstore>`_, you should *not* need this. If
        you provide a password here, it will be passed on to the
        :py:func:`extract_zip` function.
    :param skip_img_formats: The image formats to skip when attempting to
        convert them to PNG. This will typically include the strings ICO, PNG,
        and WEBP.
    :type skip_img_formats: list or tuple
    :param bool unpack_in_subprocess: Flag indicating if the job of unpacking
        the CRX should be done in a subprocess rather than calling the function
        directly. Usually this shouldn't need to be set as it will only hinder
        performance.
    :param bool convert_in_subprocess: Flag indicating if the job of converting
        the images in the CRX should be done in a subprocess rather than calling
        the function directly. Usually this SHOULD be set, since converting
        images can sometimes cause a segmentation fault, which kills the whole
        process.
    :param bool do_convert: Flag indicating whether images should be converted
        during the unpacking process (intended to mimic Chrome's unpacking
        process more closely).
    :return: Directory where the archive was extracted.
    :rtype: str
    """
    if img_tallies is None:
        # This means that the calling function won't have access to these numbers, but for consistency's sake we'll
        # store them anyway.
        img_tallies = {}

    if skip_img_formats is None:
        skip_img_formats = []

    # Make sure the file exists, get basic info about it
    crx_file = path.abspath(crx_file)
    crx_dir, base = path.split(crx_file)
    crx_size = path.getsize(crx_file)  # Raises an error for us if the file doesn't exist
    if not re.search('\.crx$', base):
        raise OSError('File has unsupported extension, expected ".crx"')

    zip_path, signature, pub_key = [None] * 3
    with open(crx_file, 'rb') as fin:
        header_vals = fin.read(4 * 4)  # 4 values, each 4 bytes (32 bits) long
        if len(header_vals) < 16:
            raise BadCrxHeader('Invalid header length')
        magic, version, pup_key_len, sig_len = HEADER_FMT.unpack(header_vals)
        if magic != b'Cr24':
            raise BadCrxHeader('Invalid magic number: %s' % codecs.encode(magic, 'hex').decode('utf-8'))
        if version != 2:
            raise BadCrxHeader('Invalid version number: %d' % version)

        # Read in the public key and signature
        pub_key = fin.read(pup_key_len)
        signature = fin.read(sig_len)

        # TODO: Add verification methods for the public key and the signature
        # verify_pub_key(pub_key)
        # verify_signature(signature)

        # Detach zip file
        with NamedTemporaryFile('wb', suffix='.zip', delete=False) as fout:
            zip_path = fout.name
            logging.debug('Created a named temp file at: {}'.format(zip_path))
            # Read the rest of the file and save it as a .zip
            fout.write(fin.read())

    if None in (zip_path, signature, pub_key):
        raise IOError('Could not separate zip file from the CRX.')

    # Extract the zip file
    path.getsize(zip_path)
    if ext_dir is None:
        ext_dir = path.join(crx_dir, base.rsplit('.', 1)[0])
    ext_dir = path.abspath(ext_dir)

    if path.isdir(ext_dir):
        if overwrite_if_exists:
            # Delete the entire directory and its contents. Ignore errors because the files will likely
            # be overwritten upon unzip anyway.
            rmtree(ext_dir, ignore_errors=True)
        else:
            err = FileExistsError()
            err.errno = ''
            err.strerror = 'Cannot unpack CRX to directory that already exists'
            err.filename = ext_dir
            raise err

    if unpack_in_subprocess:
        prog = ['python3', __file__]
        if not test_contents:
            prog.append('-t')
        prog += ['xo', zip_path, ext_dir]
        if passwd is not None:
            prog.append(passwd)
        try:
            check_call(prog)
        except CalledProcessError as err:
            if 0 < err.returncode < len(ERR_TYPE):
                e = ERR_TYPE[err.returncode]()
                logging.warning('Got error of type "%s" while unpacking file at: %s' %
                                (e.__class__.__name__, ext_dir))
                # Re-raise the original exception
                raise ERR_TYPE[err.returncode]
            logging.warning('Got error of unknown type. Return code was: %d' % err.returncode)
            raise
    else:
        extract_zip(zip_path, ext_dir, pwd=passwd, test_contents=test_contents)

    if do_convert and not FORBID_IMG_CONVERT:
        if convert_in_subprocess:
            # Get the logger info so the subprocess can recreate it
            log_obj = logging.getLogger()
            fmt = log_obj.handlers[0].formatter._fmt
            log_file = log_obj.handlers[0].baseFilename
            level = log_obj.level

            # Create the subprocess
            prog = ['python3', __file__, 'convert', ext_dir, '--log-file=%s' % log_file, '--log-level=%s' % level,
                    '--log-fmt=%s' % fmt]
            for f in skip_img_formats:
                prog += ['-s', f]
            try:
                check_call(prog)
            except CalledProcessError:
                logging.warning('Image conversion subprocess failed while unpacking  %s' % crx_file)
                with open('failed_conversions.txt', 'a') as fout:
                    fout.write(crx_file + '\n')
        else:
            convert_imgs(ext_dir, img_tallies=img_tallies, skip_other=skip_img_formats)

    set_mode(ext_dir)  # Set mode after converting in case PIL changes things
    return ext_dir


def extract_zip(zip_file, extract_dir, pwd=None, test_contents=True, reraise_errors=not JUST_EXIT):
    """Simple wrapper around the Python zipfile.ZipFile class.

    Typically, it is not necessary to call this function directly from
    anywhere other than the :py:func:`unpack` function.

    :param str zip_file: Path to the zip file to be extracted.
    :param str extract_dir: Directory where the contents will be extracted.
    :param str pwd: Password for the zip file.
    :param bool test_contents: Whether to use the library's testzip() function
        on the archive before extracting. Tests if the CRC and header of each
        file in the archive are valid.
    :param bool reraise_errors: Set to False when the ``unpack`` script is run
        with the `xo` (extract only) command, in which case the function will
        return a non-zero value when an error occurs. The default, False,
        indicates that any errors that come up should just be re-raised.
    :rtype: None
    """
    try:
        zip_obj = ZipFile(zip_file)
        if test_contents and zip_obj.testzip() is not None:
            # A file's CRC and/or header was invalid
            raise BadZipFile
        zip_obj.extractall(extract_dir, pwd=pwd)
    except ERR_TYPE[1:] as err:
        if reraise_errors:
            raise
        for i in range(1, len(ERR_TYPE)):
            if isinstance(err, ERR_TYPE[i]):
                exit(i)


def set_mode(base_dir, file_mode=FILE_MODE, dir_mode=DIR_MODE):
    """Set file and dir permissions for everything under base_dir.

    :param str base_dir: Top directory where to start working on changing the
        file and dir modes.
    :param int file_mode: The permissions number to give all files in octal. The
        default is what Chrome OS uses on files.
    :param int dir_mode: The permissions number to give all dirs in octal. The
        default is what Chrome OS uses on dirs.
    :rtype: None
    """
    # Verify we're running in POSIX system first. No need to do this if we're in Windows.
    if os.name != 'posix':
        return

    # These are the file and dir permissions to set. File: 644  Dir: 700
    file_mode = _mode_from_num(file_mode)
    dir_mode = _mode_from_num(dir_mode)
    for root, dirs, files in os.walk(base_dir):
        for name in files:
            os.chmod(path.join(root, name), file_mode)
        for name in dirs:
            os.chmod(path.join(root, name), dir_mode)


def _mode_from_num(num):
    """Return the ORed stat objects representing the octal number num.

    :param int num: Permissions number in octal, e.g. 644.
    :return: The equivalent of bitwise ORing the permission constants in the
        stat library.
    :rtype: int
    """
    assert num > 100  # The user should at least be able to read the file...

    usr = int(num / 100)
    grp = int(num / 10) - usr * 10
    oth = num % 10

    return usr << 6 | grp << 3 | oth


def convert_imgs(base_dir, skip_gifs=True, skip_other=None, img_tallies=None):
    """Convert all images under base_dir to PNG format.

    Just like Chrome, the file extension remains unchanged. Also, GIFs are
    skipped to preserve their animations if skip_gifs is True.

    :param str base_dir: The directory to walk through.
    :param bool skip_gifs: When True, GIFs won't be converted to preserve their
        animations.
    :param skip_other: The image formats to skip when attempting to convert
        them to PNG. This will typically include the strings ICO, PNG, and
        WEBP.
    :type skip_other: list|tuple
    :param dict img_tallies: A dictionary for storing the number of each type of
        image file converted during the unpacking process.
    :rtype: None
    """
    if skip_other is None:
        skip_other = []

    if img_tallies is None:
        # This means that the calling function won't have access to these numbers, but for consistency's sake we'll
        # store them anyway.
        img_tallies = {}

    for root, dirs, files in os.walk(base_dir):
        for name in files:
            fname = path.join(root, name)

            # Check that it the file has a non-zero size
            if not path.getsize(fname):
                continue

            try:
                img = Image.open(fname)
            except OSError:
                # Means the file isn't an image
                pass
            except:
                logging.warning('Got unhandled exception during image conversion of file: %s' % fname, exc_info=1)
            else:
                # Increase the tally for this image type
                f = img.format
                if f not in img_tallies.keys():
                    img_tallies[f] = 0
                img_tallies[f] += 1

                # Don't attempt to convert certain types of images
                if skip_gifs and f == 'GIF':
                    continue
                if f in skip_other:
                    continue
                try:
                    # The save will fail in certain cases if the image isn't converted to RGBA mode, which is a
                    # normal RGB mode but with transparency. The palette of 'WEB' is a guess, but seemed a better
                    # option than the other one available for that function.
                    img.convert(mode='RGBA', palette='WEB').save(fname, format='PNG')
                except OSError:
                    # Means the file isn't an image or has no length
                    pass
                except:
                    logging.warning('Got unhandled exception while SAVING a converted image: %s' % fname, exc_info=1)


def verify_pub_key(pub_key):
    raise NotImplementedError


def verify_signature(sig):
    raise NotImplementedError



# *-* coding: utf-8 *-*
"""
The :py:class:`EncryptedTempDirectory` class is designed for when you need a
directory whose contents are encrypted by `eCryptfs <http://ecryptfs.org/>`_,
but you would also like to take advantage of the features of in-memory-only
directories, such as increased access speed and automatic deletion (e.g.\  when
you're :doc:`unpacking a CRX <crx_unpack>`).

To understand this class fully, please also read the documentation on the
:py:func:`~tempfile.TemporaryDirectory` class.
"""

from base64 import b64encode
from os import mkdir, remove
from os.path import basename, isdir, join, expanduser
from re import search
from subprocess import run, PIPE
from tempfile import TemporaryDirectory


class EncryptedTempDirectory(TemporaryDirectory):
    """Create and return an encrypted temporary directory.

    This behaves similarly to :class:`TemporaryDirectory`, except for the
    following:

    - It requires that an "upper directory" be specified, which will be the
      mount point used by eCryptfs to mount the encrypted directory to the
      filesystem.
    - It creates two files in `~/.ecryptfs` required to mount the directory
      (both of which are deleted when this object is):

      - `ALIAS.sig` - Contains the signatures for the FEK and FNEK encryption
        keys.
      - `ALIAS.conf` - Contains `fstab`-style information for which directory
        eCryptfs should mount and where.

    In the above notes, `ALIAS` (which is a term used in the eCryptfs
    documentation, see links below) will be the name of the created temp
    directory, accessible as the basename of ``self.name``.

    To use an :class:`EncryptedTempDirectory` object, it's best to use it with
    a ``with`` clause, like so:

    ::

        with EncryptedTempDirectory(upper_dir=upper) as lower:
            ...

    Better yet, use an instance of :class:`TemporaryDirectory` as the upper
    directory, like this:

    ::

        with TemporaryDirectory() as upper, \\
                EncryptedTempDirectory(upper_dir=upper) as lower:
            ...

    .. note::

        In the above example, both temporary directories are deleted as soon as
        the ``__exit__()`` method is called (triggered by the close of the
        ``with`` clause). So make sure that anything you need to do with these
        objects, you do before leaving the ``with`` clause.

    .. note::

        This class depends on eCryptfs, so it will need to be installed on the
        system to work properly. Similarly, this class depends on the
        following Unix tools/devices:

        - ``head``
        - ``ecryptfs-add-passphrase``
        - ``mount``
        - ``keyctl``
        - ``/dev/urandom``

        On Debian/Ubuntu-based systems, you can install these with:

        ::

            sudo apt-get install coreutils mount keyutils ecryptfs-utils

    For more information, see the following resources:

    - http://manpages.ubuntu.com/manpages/zesty/en/man1/mount.ecryptfs_private.1.html
    - http://manpages.ubuntu.com/manpages/zesty/en/man1/ecryptfs-add-passphrase.1.html
    - https://askubuntu.com/questions/574110/how-to-use-ecryptfs-with-a-random-directory/574425#574425
    """

    def __init__(self, *, upper_dir, **kwargs):
        """
        :param str upper_dir: Path where the encrypted directory will be
            mounted, and where the unencrypted version of the files will be
            accessible.
        :param kwargs: Additional parameters to pass to the constructor of the
            :py:class:`~tempfile.TemporaryDirectory` class.
        """
        self._upper_dir = upper_dir
        super().__init__(**kwargs)

        # Make the ~/.ecryptfs dir if it doesn't already exist
        ecrypt_dir = expanduser(join('~', '.ecryptfs'))
        if not isdir(ecrypt_dir):
            mkdir(ecrypt_dir, mode=0o700)

        # Prep dirs for eCryptfs
        self._alias = basename(self.name)
        self._conf_file = '{}.conf'.format(join(ecrypt_dir, self._alias))
        with open(self._conf_file, 'w') as conf:
            conf.write('{} {} ecryptfs'.format(self.name, self._upper_dir))

        # Create the signature file and add it to the keyring
        self._sig_file = '{}.sig'.format(join(ecrypt_dir, self._alias))
        passwd = b64encode(run(['head',  '-c 32', '/dev/urandom'], check=True, stdout=PIPE).stdout)
        sig = run(['ecryptfs-add-passphrase', '--fnek'], input=passwd, stdout=PIPE).stdout.decode('utf-8')
        with open(self._sig_file, 'w') as sig_out:
            for line in sig.split('\n'):
                # There should be two matches. We want both in the signature file.
                m = search(r'sig \[(.*?)\]', line)
                if m:
                    sig_out.write(m.group(1))
                    sig_out.write('\n')

    def __enter__(self):
        # Mount the encrypted dir
        try:
            run(['mount.ecryptfs_private', self._alias], check=True)
        except:
            # No matter what the exception, we should try to remove the keys from the keyring
            keys = {}
            # Get a list of the keys so we can reference them later
            key_output = run(['keyctl', 'show'], stdout=PIPE).stdout.decode('utf-8')
            for key_line in key_output.split('\n'):
                m = search(r'^ *([0-9]*).* ([a-f0-9]*)$', key_line)
                if m:
                    keys[m.group(2)] = m.group(1)
            # The contents of the sig file will help us determine which keys should be deleted
            with open(self._sig_file) as sig_file:
                for line in sig_file:
                    line = line.strip()
                    if not len(line):
                        continue
                    # Unlink the key
                    run(['keyctl', 'unlink', keys[line]])

            self.cleanup()  # Do all the regular cleanup
            raise  # Re-raise the original exception

        return super().__enter__()

    def __exit__(self, exc, value, tb):
        # Unmount the encrypted dir
        run(['umount.ecryptfs_private', self._alias], check=True)
        super().__exit__(exc, value, tb)

    def cleanup(self):
        # Remove all the working files
        for f in (self._conf_file, self._sig_file):
            remove(f)
        super().cleanup()



#!/usr/bin/env python3
# *-* coding: utf-8 *-*

from os import path
from setuptools import setup, find_packages

with open(path.join('crx_unpack', 'VERSION')) as v_file:
    version = v_file.read().strip()


def read(fname):
    return open(path.join(path.dirname(__file__), fname)).read()

setup(
    name='crx_unpack',
    packages=find_packages(exclude=['docs', ]),
    package_data={'crx_unpack': ['VERSION']},
    scripts=['unpack'],
    version=version,
    license='MIT',
    description='Unpack .crx files the way Chrome does',
    long_description=read('README.rst'),
    author='Mike Mabey',
    author_email='mmabey@ieee.org',
    url='http://crx-unpack.readthedocs.io/',
    download_url='https://bitbucket.org/mmabey/crx_unpack/get/{}.tar.gz'.format(version),
    install_requires=[x.strip() for x in open(path.join(path.dirname(__file__), 'requirements.txt'))],
    keywords=[
        'crx',
        'unpack',
        'Chrome',
        'Chrome OS',
        'extension',
        'package',
    ],
    classifiers=[  # Full list at: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP :: Browsers',
        'Topic :: System :: Archiving :: Compression',
        'Topic :: Multimedia :: Graphics :: Graphics Conversion',
        # Operating systems supported
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS',
        # Versions of Python supported
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)
